import pytest

import tonio
import tonio.sync
import tonio.sync.channel as channel


class AtomicError(RuntimeError): ...


def test_semaphore(run):
    stack = []

    def _count(semaphore, i):
        with (yield semaphore()):
            stack.append(i)
            if len(stack) > 2:
                raise AtomicError
            yield
            stack.pop(0)
        return i

    def _run(value):
        semaphore = tonio.sync.Semaphore(value)
        out = yield tonio.spawn(*[_count(semaphore, i) for i in range(50)])
        return out

    assert run(_run(2)) == list(range(50))

    with pytest.raises(ExceptionGroup):
        run(_run(50))


def test_lock(run):
    stack = []

    def _count(lock, i):
        with (yield lock()):
            stack.append(i)
            if len(stack) > 1:
                raise AtomicError
            yield
            stack.pop(0)
        return i

    def _run():
        lock = tonio.sync.Lock()
        out = yield tonio.spawn(*[_count(lock, i) for i in range(50)])
        return out

    assert run(_run()) == list(range(50))


def test_barrier(run):
    barrier = tonio.sync.Barrier(3)
    stack = []

    def _wait():
        stack.append(True)
        i = yield barrier.wait()
        assert len(stack) == 3
        return i

    def _run():
        out = yield tonio.spawn(*[_wait() for _ in range(3)])
        return out

    assert set(run(_run())) == {0, 1, 2}


def test_channel(run):
    def _produce(sender, barrier, offset, no):
        for i in range(no):
            message = offset + i
            yield sender.send(message)
        yield barrier.wait()

    def _consume(receiver):
        messages = []
        while True:
            try:
                message = yield receiver.receive()
                messages.append(message)
            except Exception:
                break
        return messages

    def _close(sender, barrier):
        yield barrier.wait()
        sender.close()

    def _run2p4c():
        sender, receiver = channel.channel(2)
        barrier = tonio.sync.Barrier(3)
        tasks = [
            _produce(sender, barrier, 100, 20),
            _produce(sender, barrier, 200, 20),
            _consume(receiver),
            _consume(receiver),
            _consume(receiver),
            _consume(receiver),
            _close(sender, barrier),
        ]
        [_, _, c1, c2, c3, c4, _] = yield tonio.spawn(*tasks)
        return c1, c2, c3, c4

    def _run4p2c():
        sender, receiver = channel.channel(2)
        barrier = tonio.sync.Barrier(5)
        tasks = [
            _produce(sender, barrier, 100, 10),
            _produce(sender, barrier, 200, 10),
            _produce(sender, barrier, 300, 10),
            _produce(sender, barrier, 400, 10),
            _consume(receiver),
            _consume(receiver),
            _close(sender, barrier),
        ]
        [_, _, _, _, c1, c2, _] = yield tonio.spawn(*tasks)
        return c1, c2

    consumed = run(_run2p4c())
    consumed = {v for c in consumed for v in c}
    assert len(consumed) == 40
    assert consumed == ({*range(100, 120)} | {*range(200, 220)})

    consumed = run(_run4p2c())
    consumed = {v for c in consumed for v in c}
    assert len(consumed) == 40
    assert consumed == ({*range(100, 110)} | {*range(200, 210)} | {*range(300, 310)} | {*range(400, 410)})


def test_channel_unbounded(run):
    def _produce(sender, barrier, offset, no):
        for i in range(no):
            message = offset + i
            sender.send(message)
        yield barrier.wait()

    def _consume(receiver):
        messages = []
        while True:
            try:
                message = yield receiver.receive()
                messages.append(message)
            except Exception:
                break
        return messages

    def _close(sender, barrier):
        yield barrier.wait()
        sender.close()

    def _run2p4c():
        sender, receiver = channel.unbounded()
        barrier = tonio.sync.Barrier(3)
        tasks = [
            _produce(sender, barrier, 100, 20),
            _produce(sender, barrier, 200, 20),
            _consume(receiver),
            _consume(receiver),
            _consume(receiver),
            _consume(receiver),
            _close(sender, barrier),
        ]
        [_, _, c1, c2, c3, c4, _] = yield tonio.spawn(*tasks)
        return c1, c2, c3, c4

    def _run4p2c():
        sender, receiver = channel.unbounded()
        barrier = tonio.sync.Barrier(5)
        tasks = [
            _produce(sender, barrier, 100, 10),
            _produce(sender, barrier, 200, 10),
            _produce(sender, barrier, 300, 10),
            _produce(sender, barrier, 400, 10),
            _consume(receiver),
            _consume(receiver),
            _close(sender, barrier),
        ]
        [_, _, _, _, c1, c2, _] = yield tonio.spawn(*tasks)
        return c1, c2

    consumed = run(_run2p4c())
    consumed = {v for c in consumed for v in c}
    assert len(consumed) == 40
    assert consumed == ({*range(100, 120)} | {*range(200, 220)})

    consumed = run(_run4p2c())
    consumed = {v for c in consumed for v in c}
    assert len(consumed) == 40
    assert consumed == ({*range(100, 110)} | {*range(200, 210)} | {*range(300, 310)} | {*range(400, 410)})



import contextlib
import threading
from typing import Any, Awaitable, Callable, Iterable, ParamSpec, TypeVar

from .._tonio import CancelledError, ResultHolder, get_runtime
from ._events import Event
from ._sync import Barrier


_T = TypeVar('_T')
_Params = ParamSpec('_Params')
_Return = TypeVar('_Return')


class _SpawnJoinCollect:
    __slots__ = ['_barrier', '_res', '_errs']

    def __init__(self, barrier, res, errs):
        self._barrier = barrier
        self._res = res
        self._errs = errs

    def __await__(self):
        return self._wait().__await__()

    async def _wait(self):
        await self._barrier.wait()
        if self._errs:
            raise ExceptionGroup('SpawnExceptionGroup', self._errs)
        return self._res.fetch()


class _SpawnJoin:
    __slots__ = ['_barrier', '_errs']

    def __init__(self, barrier, errs):
        self._barrier = barrier
        self._errs = errs

    def __await__(self):
        return self._wait().__await__()

    async def _wait(self):
        await self._barrier.wait()
        if self._errs:
            raise ExceptionGroup('SpawnExceptionGroup', self._errs)


class _Spawn:
    __slots__ = []

    @staticmethod
    def __call__(*coros) -> Awaitable[Any]:
        barrier = Barrier(len(coros) + 1)
        res = ResultHolder(len(coros))
        errs = []

        async def wrapper(idx, coro, barrier):
            try:
                ret = await coro
                res.store(ret, idx)
            except Exception as exc:
                errs.append(exc)
            finally:
                barrier.ack()

        for idx, coro in enumerate(coros):
            get_runtime()._spawn_pyasyncgen(wrapper(idx, coro, barrier))

        return _SpawnJoinCollect(barrier, res, errs)

    @staticmethod
    def without_results(*coros) -> Awaitable[None]:
        barrier = Barrier(len(coros) + 1)
        errs = []

        async def wrapper(coro, barrier):
            try:
                await coro
            except Exception as exc:
                errs.append(exc)
            finally:
                barrier.ack()

        for coro in coros:
            get_runtime()._spawn_pyasyncgen(wrapper(coro, barrier))

        return _SpawnJoin(barrier, errs)

    @staticmethod
    def without_tracking(*coros):
        for coro in coros:
            get_runtime()._spawn_pyasyncgen(coro)


spawn = _Spawn()


async def spawn_blocking(fn: Callable[_Params, _Return], /, *args: _Params.args, **kwargs: _Params.kwargs) -> _Return:
    ctl, event, res = get_runtime()._spawn_blocking(fn, *args, **kwargs)
    with contextlib.suppress(CancelledError):
        await event.waiter(None)
    err, val = res.fetch()
    if err is None:
        ctl.abort()
        await event.waiter(None)
        err, val = res.fetch()
    if err is True:
        raise val
    return val


def block_on(coro):
    ev = threading.Event()
    res = ResultHolder()

    async def wrapper():
        try:
            ret = await coro
            res.store((False, ret))
        except Exception as exc:
            res.store((True, exc))
        finally:
            ev.set()

    spawn.without_tracking(wrapper())
    ev.wait()
    err, val = res.fetch()
    if err:
        raise val
    return val


def map(fn: Callable[[_T], _Return], /, xs: Iterable[_T]) -> Awaitable[list[_Return]]:
    return spawn(*[fn(x) for x in xs])


def map_blocking(fn: Callable[[_T], _Return], /, xs: Iterable[_T]) -> Awaitable[list[_Return]]:
    return spawn(*[spawn_blocking(fn, x) for x in xs])


def yield_now() -> Awaitable[None]:
    event = Event()
    event.set()
    return event.waiter(None)



from socket import SocketType as _SocketType
from typing import Any, TypeVar

_T = TypeVar('_T')

__version__: str

class Runtime:
    _clock: int
    _closed: bool
    _sigset: list[int]
    _sig_wfd: Any
    _sig_listening: bool
    _ssock_r: Any
    _ssock_w: Any
    _stopping: bool

    def __init__(
        self, threads: int, threads_blocking: int, threads_blocking_timeout: int, context: bool, signals: list[int]
    ): ...
    def _run(self): ...
    def _spawn_pygen(self, coro): ...
    def _spawn_pyasyncgen(self, coro): ...
    def _spawn_blocking(self, f, *args, **kwargs) -> tuple[BlockingTaskCtl, Event, ResultHolder]: ...
    def _io_event_r(self, fd: int) -> Event: ...
    def _io_event_w(self, fd: int) -> Event: ...
    def _sig_add(self, sig: int) -> Event: ...
    def _sig_rem(self, sig: int) -> bool: ...

class Waiter:
    def __init__(self, *events: Event): ...

class Event:
    def __init__(self): ...
    def set(self): ...
    def clear(self): ...
    def is_set(self) -> bool: ...
    def waiter(self, timeout: int | None) -> Waiter: ...

class ResultHolder:
    def __init__(self, size: int = 1): ...
    def store(self, value, index: int | None = None): ...
    def fetch(self): ...

class Scope:
    def __init__(self): ...
    def _incr(self, val: int) -> bool: ...
    def _track_pygen(self, coro) -> Event: ...
    def _track_pyasyncgen(self, coro) -> Event: ...
    def _stack(self) -> tuple[Waiter, list[Any]]: ...
    def cancel(self) -> bool: ...

class BlockingTaskCtl:
    def abort(self): ...

class Lock:
    def acquire(self) -> None | Event: ...
    def try_acquire(self): ...
    def release(self): ...

class LockCtx:
    def __init__(self, lock: Lock): ...
    def __enter__(self): ...
    def __exit__(self, exc_type, exc_value, exc_tb): ...

class Semaphore:
    def __init__(self, value: int): ...
    def acquire(self) -> None | Event: ...
    def try_acquire(self): ...
    def release(self): ...
    def tokens(self) -> int: ...

class SemaphoreCtx:
    def __init__(self, semaphore: Semaphore): ...
    def __enter__(self): ...
    def __exit__(self, exc_type, exc_value, exc_tb): ...

class Barrier:
    _event: Event

    def __init__(self, value: int): ...
    def ack(self) -> int: ...
    def value(self) -> int: ...

class Channel:
    def __init__(self, size: int): ...

class ChannelSender:
    def __init__(self, channel: Channel): ...
    def _send(self, message) -> Event: ...
    def close(self): ...

class ChannelReceiver[_T]:
    def __init__(self, channel: Channel): ...
    def _receive(self) -> tuple[Event, bool, _T]: ...

class UnboundedChannel:
    def __init__(self): ...

class UnboundedChannelSender[_T]:
    def __init__(self, channel: UnboundedChannel): ...
    def send(self, message: _T): ...

class UnboundedChannelReceiver[_T]:
    def __init__(self, channel: UnboundedChannel): ...
    def _receive(self) -> tuple[Event, bool, _T]: ...

class CancelledError(BaseException): ...
class ResourceBroken(Exception): ...
class RuntimeAlreadyInitializedError(RuntimeError): ...
class RuntimeNotInitializedError(RuntimeError): ...
class TimeoutError(BaseException): ...
class WouldBlock(Exception): ...

class Socket:
    _sock: _SocketType

    def __init__(self, stdlib_socket: _SocketType): ...
    def _eof_get(self) -> bool: ...
    def _eof_set(self): ...

class TLSStream:
    _state: int

    def _handshake_pre(self): ...
    def _handshake_post(self): ...
    def _set_broken(self): ...
    def _set_closed(self): ...
    def _check_ready(self): ...

def get_runtime() -> Runtime: ...
def set_runtime(runtime: Runtime): ...



"""
Heavily inspired by `trio` code.

:source: (https://github.com/python-trio/trio)
:copyright: Contributors to the Trio project
:license: MIT
"""

import contextlib
import ssl as _stdlib_ssl
from typing import Any

from ..._net._tls import _is_eof
from ..._tonio import ResourceBroken, TLSStream as _TLSStream
from .._sync import Lock
from ._streams import _Stream


class TLSStream(_Stream, _TLSStream):
    __slots__ = [
        'transport',
        '_ssl',
        '_egress',
        '_ingress',
        '_lock_recv',
        '_lock_send',
        '_egress_stack',
        '_recv_count',
        '_recv_est_size',
        '_compat_https',
    ]

    def __init__(
        self,
        transport: _Stream,
        ssl_context: _stdlib_ssl.SSLContext,
        *,
        server_hostname: str | bytes | None = None,
        server_side: bool = False,
        https_compatible: bool = False,
    ):
        self.transport = transport
        self._compat_https = https_compatible
        self._egress = _stdlib_ssl.MemoryBIO()
        self._ingress = _stdlib_ssl.MemoryBIO()
        self._lock_recv = Lock()
        self._lock_send = Lock()
        self._egress_stack = bytearray()
        self._recv_count = 0
        self._recv_est_size = 16384
        self._ssl = ssl_context.wrap_bio(
            self._ingress,
            self._egress,
            server_side=server_side,
            server_hostname=server_hostname,
        )

    async def _recv(self) -> None:
        recv_count = self._recv_count
        async with self._lock_recv:
            if recv_count == self._recv_count:
                data = await self.transport.receive_some()
                if not data:
                    self._ingress.write_eof()
                else:
                    self._recv_est_size = max(
                        self._recv_est_size,
                        len(data),
                    )
                    self._ingress.write(data)
                self._recv_count += 1

    async def _send(self, data) -> None:
        async with self._lock_send:
            try:
                self._egress_stack.extend(data)
                data = bytes(self._egress_stack)
                self._egress_stack.clear()
                await self.transport.send_all(data)
            except:
                self._set_broken()
                raise

    async def _ssl_dance(self, ssl_fn, *args) -> Any:
        done = False
        while not done:
            want_read = False
            try:
                ret = ssl_fn(*args)
            except _stdlib_ssl.SSLWantReadError:
                want_read = True
            except (_stdlib_ssl.SSLError, _stdlib_ssl.CertificateError) as exc:
                self._set_broken()
                raise ResourceBroken from exc
            else:
                done = True

            if to_send := self._egress.read():
                await self._send(to_send)
            elif want_read:
                await self._recv()

        return ret

    async def handshake(self) -> None:
        self._handshake_pre()

        done = False
        while not done:
            want_read = False
            try:
                self._ssl.do_handshake()
            except _stdlib_ssl.SSLWantReadError:
                want_read = True
            except (_stdlib_ssl.SSLError, _stdlib_ssl.CertificateError) as exc:
                self._set_broken()
                raise ResourceBroken from exc
            else:
                done = True

            to_send = self._egress.read()
            if not want_read and self._ssl.server_side and self._ssl.version() == 'TLSv1.3':
                self._egress_stack.extend(to_send)
                to_send = b''

            if to_send:
                await self._send(to_send)
            elif want_read:
                await self._recv()

        self._handshake_post()

    async def send_all(self, data: bytes | bytearray | memoryview) -> None:
        self._check_ready()
        if not data:
            return
        await self._ssl_dance(self._ssl.write, data)

    async def receive_some(self, max_bytes: int | None = None) -> bytes | bytearray:
        self._check_ready()
        if max_bytes is None:
            max_bytes = max(self._recv_est_size, self._ingress.pending)
        try:
            ret = await self._ssl_dance(self._ssl.read, max_bytes)
            return ret
        except ResourceBroken as exc:
            if self._compat_https and _is_eof(exc.__cause__):
                return b''
            raise

    async def close(self) -> None:
        if self._state == 4:
            return

        self._set_closed()
        if self._state == 3 or self._compat_https:
            self.transport.close()
            return

        try:
            with contextlib.suppress(ResourceBroken):
                done = False
                while not done:
                    try:
                        self._ssl.unwrap()
                    except _stdlib_ssl.SSLWantReadError:
                        done = True
                    except (_stdlib_ssl.SSLError, _stdlib_ssl.CertificateError) as exc:
                        self._set_broken()
                        raise ResourceBroken from exc
                    else:
                        done = True

                    if to_send := self._egress.read():
                        await self._send(to_send)
        finally:
            self.transport.close()

    def __exit__(self, *args, **kwargs) -> None:
        if self._state == 4:
            return

        if self._state == 3 or self._compat_https:
            self._set_closed()
            self.transport.close()
            return

        raise RuntimeError('TLSStream needs to be manually closed')


class TLSListener(_Stream):
    __slots__ = [
        'transport',
        '_ssl',
        '_compat_https',
    ]

    def __init__(
        self,
        transport,
        ssl_context: _stdlib_ssl.SSLContext,
        *,
        https_compatible: bool = False,
    ) -> None:
        self.transport = transport
        self._ssl = ssl_context
        self._compat_https = https_compatible

    async def accept(self) -> TLSStream:
        stream = await self.transport.accept()
        ret = TLSStream(
            stream,
            self._ssl,
            server_side=True,
            https_compatible=self._compat_https,
        )
        await ret.handshake()
        return ret

    def close(self) -> None:
        self.transport.close()



import tonio
from tonio.net import SocketStream, open_tcp_stream, serve_tcp, socket


_SIZE = 1024 * 1024


def _get_port():
    sock = socket.socket()

    with sock:
        yield sock.bind(('127.0.0.1', 0))
        name = sock.getsockname()
        return name[1]


def test_streams_tcp_recv(run):
    def server():
        done = tonio.Event()
        res = []
        port = yield _get_port()

        def _server_handler(stream: SocketStream):
            buf = b''
            while len(buf) < _SIZE:
                buf += yield stream.receive_some()
            res.append(buf)
            done.set()

        with tonio.scope() as scope:
            scope.spawn(serve_tcp(_server_handler, host='127.0.0.1', port=port))
            scope.spawn(client(port))
            yield done.wait()
            scope.cancel()
        yield scope()

        return res[0]

    def client(port):
        yield tonio.sleep(0.5)
        stream: SocketStream = yield open_tcp_stream('127.0.0.1', port=port)
        yield stream.send_all(b'a' * _SIZE)

    data = run(server())
    assert data == b'a' * _SIZE


def test_streams_tcp_send(run):
    done = tonio.Event()
    state = {'data': b''}

    def server():
        port = yield _get_port()

        def _server_handler(stream: SocketStream):
            yield stream.send_all(b'a' * _SIZE)
            stream.send_eof()

        with tonio.scope() as scope:
            scope.spawn(serve_tcp(_server_handler, host='127.0.0.1', port=port))
            scope.spawn(client(port))
            yield done.wait()
            scope.cancel()
        yield scope()

    def client(port):
        yield tonio.sleep(0.5)
        stream: SocketStream = yield open_tcp_stream('127.0.0.1', port=port)
        while len(state['data']) < _SIZE:
            state['data'] += yield stream.receive_some()
        done.set()

    run(server())
    assert state['data'] == b'a' * _SIZE



import contextlib
import threading
from typing import Any, Callable, Iterable, ParamSpec, TypeVar

from ._sync import Barrier
from ._tonio import CancelledError, ResultHolder, get_runtime
from ._types import Coro


_T = TypeVar('_T')
_Params = ParamSpec('_Params')
_Return = TypeVar('_Return')


class _Spawn:
    __slots__ = []

    @staticmethod
    def __call__(*coros: Coro) -> Coro[Any]:
        barrier = Barrier(len(coros) + 1)
        res = ResultHolder(len(coros))
        errs = []

        def wrapper(idx, coro, barrier):
            try:
                ret = yield coro
                res.store(ret, idx)
            except Exception as exc:
                errs.append(exc)
            finally:
                barrier.ack()

        for idx, coro in enumerate(coros):
            get_runtime()._spawn_pygen(wrapper(idx, coro, barrier))

        def join():
            yield barrier.wait()
            if errs:
                raise ExceptionGroup('SpawnExceptionGroup', errs)
            return res.fetch()

        return join()

    @staticmethod
    def without_results(*coros: Coro) -> Coro[None]:
        barrier = Barrier(len(coros) + 1)
        errs = []

        def wrapper(coro, barrier):
            try:
                yield coro
            except Exception as exc:
                errs.append(exc)
            finally:
                barrier.ack()

        for coro in coros:
            get_runtime()._spawn_pygen(wrapper(coro, barrier))

        def join():
            yield barrier.wait()
            if errs:
                raise ExceptionGroup('SpawnExceptionGroup', errs)

        return join()

    @staticmethod
    def without_tracking(*coros: Coro):
        for coro in coros:
            get_runtime()._spawn_pygen(coro)


spawn = _Spawn()


def spawn_blocking(fn: Callable[_Params, _Return], /, *args: _Params.args, **kwargs: _Params.kwargs) -> Coro[_Return]:
    ctl, event, res = get_runtime()._spawn_blocking(fn, *args, **kwargs)
    with contextlib.suppress(CancelledError):
        yield event.waiter(None)
    err, val = res.fetch()
    if err is None:
        ctl.abort()
        yield event.waiter(None)
        err, val = res.fetch()
    if err is True:
        raise val
    return val


def block_on(coro: Coro[_T]) -> _T:
    ev = threading.Event()
    res = ResultHolder()

    def wrapper():
        try:
            ret = yield coro
            res.store((False, ret))
        except Exception as exc:
            res.store((True, exc))
        finally:
            ev.set()

    spawn.without_tracking(wrapper())
    ev.wait()
    err, val = res.fetch()
    if err:
        raise val
    return val


def map(fn: Callable[[_T], _Return], /, xs: Iterable[_T]) -> Coro[list[_Return]]:
    return spawn(*[fn(x) for x in xs])


def map_blocking(fn: Callable[[_T], _Return], /, xs: Iterable[_T]) -> Coro[list[_Return]]:
    return spawn(*[spawn_blocking(fn, x) for x in xs])



"""
Heavily inspired by `trio` code.

:source: (https://github.com/python-trio/trio)
:copyright: Contributors to the Trio project
:license: MIT
"""

import errno
import os
import socket as _stdlib_socket
import ssl as _stdlib_ssl
import sys
from contextlib import contextmanager
from typing import Any, Generator

from .._ctl import spawn
from .._events import Event
from .._scope import scope
from .._time import sleep
from .._types import Coro
from ._socket import _Socket, getaddrinfo, socket
from ._streams import SocketListener, SocketStream
from ._tls import TLSListener, TLSStream


_accept_retry_errnos = {
    errno.EMFILE,
    errno.ENFILE,
    errno.ENOMEM,
    errno.ENOBUFS,
}


@contextmanager
def _close_all_sockets() -> Generator[set[_Socket], None, None]:
    socks: set[_Socket] = set()
    try:
        yield socks
    finally:
        errs = []
        for sock in socks:
            try:
                sock.close()
            except BaseException as exc:
                errs.append(exc)
        if len(errs) == 1:
            raise errs[0]
        elif errs:
            raise BaseExceptionGroup('', errs)


@contextmanager
def _close_on_error(obj: Any) -> Generator[Any, None, None]:
    try:
        yield obj
    except:
        obj.close()
        raise


def open_tcp_stream(
    host: str | bytes,
    port: int,
    *,
    happy_eyeballs_delay: float | None = None,
    local_address: str | None = None,
) -> Coro[SocketStream]:
    if not isinstance(host, (str, bytes)):
        raise ValueError(f'host must be str or bytes, not {host!r}')
    if not isinstance(port, int):
        raise TypeError(f'port must be int, not {port!r}')

    if happy_eyeballs_delay is None:
        happy_eyeballs_delay = 0.250

    targets = yield getaddrinfo(host, port, type=_stdlib_socket.SOCK_STREAM)
    # TODO: IPv4, IPv6 reorder
    # reorder_targets(targets)
    errs: list[OSError] = []
    winning_socket: _Socket | None = None

    def attempt_connect(
        socket_args: tuple[int, int, int],
        sockaddr: Any,
        failed: Event,
    ):
        nonlocal winning_socket
        try:
            sock = socket(*socket_args)
            open_sockets.add(sock)

            if local_address is not None:
                # with suppress(OSError, AttributeError):
                #     sock.setsockopt(
                #         _stdlib_socket.IPPROTO_IP,
                #         _stdlib_socket.IP_BIND_ADDRESS_NO_PORT,
                #         1,
                #     )
                try:
                    yield sock.bind((local_address, 0))
                except OSError:
                    raise OSError(
                        f'local_address={local_address!r} is incompatible with remote address {sockaddr!r}',
                    ) from None

            yield sock.connect(sockaddr)
            winning_socket = sock
            _scope.cancel()
        except OSError as exc:
            errs.append(exc)
            failed.set()

    with _close_all_sockets() as open_sockets:
        with scope() as _scope:
            for address_family, socket_type, proto, _, addr in targets:
                failed = Event()
                _scope.spawn(
                    attempt_connect(
                        (address_family, socket_type, proto),
                        addr,
                        failed,
                    )
                )
                yield failed.wait(happy_eyeballs_delay)

        yield _scope()

        if winning_socket is None:
            assert len(errs) == len(targets)
            msg = f'all attempts to connect to {(host, port)} failed'
            raise OSError(msg) from ExceptionGroup(msg, errs)

        stream = SocketStream(winning_socket)
        open_sockets.remove(winning_socket)

        return stream


def open_unix_socket(filename: str | bytes | os.PathLike[str] | os.PathLike[bytes]) -> Coro[SocketStream]:
    sock = socket(_stdlib_socket.AF_UNIX, _stdlib_socket.SOCK_STREAM)
    with _close_on_error(sock):
        yield sock.connect(os.fspath(filename))
    return SocketStream(sock)


def open_tcp_listeners(
    port: int,
    *,
    host: str | bytes | None = None,
    backlog: int | None = None,
) -> Coro[list[SocketListener]]:
    if not isinstance(port, int):
        raise TypeError(f'port must be an int not {port!r}')

    backlog = min(backlog or 0xFFFF, 0xFFFF)
    addresses = yield getaddrinfo(
        host,
        port,
        type=_stdlib_socket.SOCK_STREAM,
        flags=_stdlib_socket.AI_PASSIVE,
    )

    listeners = []
    unsupported_address_families = []
    try:
        for family, type_, proto, _, sockaddr in addresses:
            try:
                sock = socket(family, type_, proto)
            except OSError as ex:
                if ex.errno == errno.EAFNOSUPPORT:
                    unsupported_address_families.append(ex)
                    continue
                else:
                    raise
            try:
                if sys.platform != 'win32':
                    sock.setsockopt(_stdlib_socket.SOL_SOCKET, _stdlib_socket.SO_REUSEADDR, 1)

                if family == _stdlib_socket.AF_INET6:
                    sock.setsockopt(_stdlib_socket.IPPROTO_IPV6, _stdlib_socket.IPV6_V6ONLY, 1)

                yield sock.bind(sockaddr)
                sock.listen(backlog)

                listeners.append(SocketListener(sock))
            except:
                sock.close()
                raise
    except:
        for listener in listeners:
            listener.close()
        raise

    if unsupported_address_families and not listeners:
        msg = "This system doesn't support any of the kinds of socket that provided address could use"
        raise OSError(errno.EAFNOSUPPORT, msg) from ExceptionGroup(
            msg,
            unsupported_address_families,
        )

    return listeners


def serve_listeners(
    handler: Any,
    listeners: list[SocketListener],
) -> Coro[None]:
    def _listener_handler(listener: SocketListener):
        with listener:
            while True:
                try:
                    stream = yield listener.accept()
                except OSError as exc:
                    if exc.errno in _accept_retry_errnos:
                        yield sleep(0.1)
                    else:
                        raise
                else:
                    spawn.without_tracking(handler(stream))

    tasks = []
    for listener in listeners:
        tasks.append(_listener_handler(listener))

    return spawn.without_results(*tasks)


def serve_tcp(
    handler: Any,
    port: int,
    *,
    host: str | bytes | None = None,
    backlog: int | None = None,
) -> Coro[None]:
    listeners = yield open_tcp_listeners(port, host=host, backlog=backlog)
    yield serve_listeners(handler, listeners)


def open_tls_over_tcp_stream(
    host: str | bytes,
    port: int,
    *,
    https_compatible: bool = False,
    ssl_context: _stdlib_ssl.SSLContext | None = None,
    happy_eyeballs_delay: float | None = None,
) -> Coro[TLSStream]:
    tcp_stream = yield open_tcp_stream(
        host,
        port,
        happy_eyeballs_delay=happy_eyeballs_delay,
    )
    if ssl_context is None:
        ssl_context = _stdlib_ssl.create_default_context()

        if hasattr(_stdlib_ssl, 'OP_IGNORE_UNEXPECTED_EOF'):
            ssl_context.options &= ~_stdlib_ssl.OP_IGNORE_UNEXPECTED_EOF

    ret = TLSStream(
        tcp_stream,
        ssl_context,
        server_hostname=host,
        https_compatible=https_compatible,
    )
    yield ret.handshake()
    return ret


def open_tls_over_tcp_listeners(
    port: int,
    ssl_context: _stdlib_ssl.SSLContext | None = None,
    *,
    host: str | bytes | None = None,
    https_compatible: bool = False,
    backlog: int | None = None,
) -> Coro[list[TLSListener]]:
    tcp_listeners = yield open_tcp_listeners(port, host=host, backlog=backlog)
    ssl_listeners = [
        TLSListener(tcp_listener, ssl_context, https_compatible=https_compatible) for tcp_listener in tcp_listeners
    ]
    return ssl_listeners


def serve_tls_over_tcp(
    handler: Any,
    port: int,
    ssl_context: _stdlib_ssl.SSLContext,
    *,
    host: str | bytes | None = None,
    https_compatible: bool = False,
    backlog: int | None = None,
) -> Coro[None]:
    listeners = yield open_tls_over_tcp_listeners(
        port,
        ssl_context,
        host=host,
        https_compatible=https_compatible,
        backlog=backlog,
    )
    yield serve_listeners(handler, listeners)



import contextlib
import inspect

from ._tonio import CancelledError, Scope as _Scope, get_runtime
from ._types import Coro


class Scope(_Scope):
    def spawn(self, coro: Coro):
        def inner(waiter):
            yield waiter
            yield coro

        def wrapper(event, waiter):
            try:
                yield inner(waiter)
            except CancelledError as exc:
                with contextlib.suppress(CancelledError):
                    waiter.throw(exc)
                while True:
                    if inspect.getgeneratorstate(coro) in [inspect.GEN_CREATED, inspect.GEN_RUNNING]:
                        yield
                        continue
                    raise coro.throw(exc)
            finally:
                event.set()

        if wrapped_coro := self._track_pygen(wrapper):
            get_runtime()._spawn_pygen(wrapped_coro)

    def __enter__(self):
        if not self._incr(0):
            raise RuntimeError('Cannot enter the same scope multiple times.')
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._incr(1)
        return

    def __call__(self):
        yield
        waiter, coros = self._stack()

        while True:
            pending = []
            for coro in coros:
                if inspect.getgeneratorstate(coro) == inspect.GEN_RUNNING:
                    pending.append(coro)
                    continue
                with contextlib.suppress(CancelledError):
                    coro.throw(CancelledError)

            if not pending:
                break
            coros = list(pending)

        yield waiter


def scope():
    return Scope()



import tonio


def test_scope_cancel(run):
    enter = []
    exit = []

    def _sleep(idx, t):
        enter.append(idx)
        yield tonio.sleep(t)
        exit.append(idx)

    def _run():
        with tonio.scope() as scope:
            scope.spawn(_sleep(1, 0.1))
            scope.spawn(_sleep(2, 2))
            yield tonio.sleep(0.2)
            scope.cancel()
        # `spawn` calls after exit are noop
        scope.spawn(_sleep(3, 0.1))

        yield scope()
        yield tonio.sleep(2)

    run(_run())

    assert set(enter) == {1, 2}
    assert set(exit) == {1}


def test_scope_cancel_immediate(run):
    enter = []
    exit = []

    def _sleep(idx, t):
        enter.append(idx)
        yield tonio.sleep(t)
        exit.append(idx)

    def _run():
        with tonio.scope() as scope:
            scope.cancel()
            scope.spawn(_sleep(1, 0.2))
        # `spawn` calls after exit are noop
        scope.spawn(_sleep(2, 0.1))

        yield scope()
        yield tonio.sleep(0.4)

    run(_run())

    assert set(enter) == {1}
    assert not exit



import time

import pytest

import tonio
import tonio.time


def test_spawn(run):
    stack = []

    def _blocking(v):
        time.sleep(0.1)
        stack.append(v)

    def _run():
        a = tonio.spawn_blocking(_blocking, 1)
        b = tonio.spawn_blocking(_blocking, 2)
        yield a
        yield b

    run(_run())
    assert set(stack) == {1, 2}


def test_err(run):
    def _blocking():
        1 / 0

    def _run():
        yield tonio.spawn_blocking(_blocking)

    with pytest.raises(ZeroDivisionError):
        run(_run())


def test_abort(run):
    stack = []

    def _blocking():
        time.sleep(3)
        stack.append(True)

    def _run():
        ret = yield tonio.time.timeout(tonio.spawn_blocking(_blocking), 1)
        yield tonio.time.sleep(3)
        return ret

    ret, completed = run(_run())
    assert not completed
    assert not ret
    assert not stack


def test_block_on(run):
    stack = []
    ev1 = tonio.Event()
    ev2 = tonio.Event()

    def _coro1():
        stack.append(1)
        ev1.set()
        yield ev2.wait()

    def _coro2():
        yield ev1.wait()
        stack.append(2)
        ev2.set()

    def _blocking():
        tonio.block_on(_coro2())

    def _run():
        a = tonio.spawn(_coro1())
        b = tonio.spawn_blocking(_blocking)
        yield b
        yield a

    run(_run())
    assert set(stack) == {1, 2}



from __future__ import annotations

import socket as _stdlib_socket
from contextlib import suppress
from types import CoroutineType
from typing import Any

from ..._net._streams import _ignorable_accept_errnos, _Stream
from ._socket import _Socket


class SocketStream(_Stream):
    __slots__ = ['socket']

    def __init__(self, socket: _Socket):
        if not isinstance(socket, _Socket):
            raise TypeError('SocketStream requires a TonIO socket object')
        if socket.type != _stdlib_socket.SOCK_STREAM:
            raise ValueError('SocketStream requires a SOCK_STREAM socket')

        self.socket = socket

        if hasattr(_stdlib_socket, 'TCP_NOTSENT_LOWAT'):
            with suppress(OSError):
                self.socket.setsockopt(_stdlib_socket.IPPROTO_TCP, _stdlib_socket.TCP_NOTSENT_LOWAT, 2**14)

    async def send_all(self, data: bytes | bytearray | memoryview) -> None:
        if self.socket._eof_get():
            raise RuntimeError("can't send data after sending EOF")
        with memoryview(data) as data:
            if not data:
                return
            total_sent = 0
            while total_sent < len(data):
                with data[total_sent:] as remaining:
                    sent = await self.socket.send(remaining)
                total_sent += sent

    def send_eof(self) -> None:
        if self.socket._eof_get():
            return
        self.socket.shutdown(_stdlib_socket.SHUT_WR)

    def receive_some(self, max_bytes: int | None = None) -> CoroutineType[Any, Any, bytes]:
        max_bytes = max_bytes or 65536
        return self.socket.recv(max_bytes)

    def close(self):
        self.socket.close()


class SocketListener(_Stream):
    def __init__(self, socket: _Socket):
        if not isinstance(socket, _Socket):
            raise TypeError('SocketStream requires a TonIO socket object')
        if socket.type != _stdlib_socket.SOCK_STREAM:
            raise ValueError('SocketStream requires a SOCK_STREAM socket')

        self.socket = socket

    async def accept(self) -> SocketStream:
        while True:
            try:
                sock, _ = await self.socket.accept()
            except OSError as exc:
                if exc.errno not in _ignorable_accept_errnos:
                    raise
            else:
                return SocketStream(sock)

    def close(self):
        self.socket.close()



from __future__ import annotations

import errno
import socket as _stdlib_socket
from abc import ABC, abstractmethod
from contextlib import suppress
from types import TracebackType

from .._types import Coro
from ._socket import _Socket


class _Stream(ABC):
    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self.close()

    @abstractmethod
    def close(self): ...


class SocketStream(_Stream):
    __slots__ = ['socket']

    def __init__(self, socket: _Socket):
        if not isinstance(socket, _Socket):
            raise TypeError('SocketStream requires a TonIO socket object')
        if socket.type != _stdlib_socket.SOCK_STREAM:
            raise ValueError('SocketStream requires a SOCK_STREAM socket')

        self.socket = socket

        if hasattr(_stdlib_socket, 'TCP_NOTSENT_LOWAT'):
            with suppress(OSError):
                self.socket.setsockopt(_stdlib_socket.IPPROTO_TCP, _stdlib_socket.TCP_NOTSENT_LOWAT, 2**14)

    def send_all(self, data: bytes | bytearray | memoryview) -> Coro[None]:
        if self.socket._eof_get():
            raise RuntimeError("can't send data after sending EOF")
        with memoryview(data) as data:
            if not data:
                return
            total_sent = 0
            while total_sent < len(data):
                with data[total_sent:] as remaining:
                    sent = yield self.socket.send(remaining)
                total_sent += sent

    def send_eof(self) -> None:
        if self.socket._eof_get():
            return
        self.socket.shutdown(_stdlib_socket.SHUT_WR)

    def receive_some(self, max_bytes: int | None = None) -> Coro[bytes]:
        max_bytes = max_bytes or 65536
        return self.socket.recv(max_bytes)

    def close(self):
        self.socket.close()


_ignorable_accept_errnos: set[int] = set()
for name in [
    'ECONNABORTED',
    'ECONNRESET',
    'EHOSTDOWN',
    'EHOSTUNREACH',
    'ENETDOWN',
    'ENETUNREACH',
    'ENONET',
    'ENOPROTOOPT',
    'ENOSR',
    'EOPNOTSUPP',
    'EPERM',
    'EPROTO',
    'EPROTONOSUPPORT',
    'ESOCKTNOSUPPORT',
    'ETIMEDOUT',
]:
    with suppress(AttributeError):
        _ignorable_accept_errnos.add(getattr(errno, name))


class SocketListener(_Stream):
    def __init__(self, socket: _Socket):
        if not isinstance(socket, _Socket):
            raise TypeError('SocketStream requires a TonIO socket object')
        if socket.type != _stdlib_socket.SOCK_STREAM:
            raise ValueError('SocketStream requires a SOCK_STREAM socket')

        self.socket = socket

    def accept(self) -> Coro[SocketStream]:
        while True:
            try:
                sock, _ = yield self.socket.accept()
            except OSError as exc:
                if exc.errno not in _ignorable_accept_errnos:
                    raise
            else:
                return SocketStream(sock)

    def close(self):
        self.socket.close()



import contextlib
import inspect

from .._tonio import CancelledError, Scope as _Scope, get_runtime
from ._ctl import yield_now


class Scope(_Scope):
    def spawn(self, coro):
        async def wrapper(event):
            try:
                await coro
            finally:
                event.set()

        if wrapped_coro := self._track_pyasyncgen(wrapper):
            get_runtime()._spawn_pyasyncgen(wrapped_coro)

    async def __aenter__(self):
        if not self._incr(0):
            raise RuntimeError('Cannot enter the same scope multiple times.')
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        self._incr(1)
        await yield_now()
        waiter, coros = self._stack()

        while True:
            pending = []
            for coro in coros:
                cstate = inspect.getcoroutinestate(coro)
                if cstate in [inspect.CORO_CREATED, inspect.CORO_RUNNING]:
                    pending.append(coro)
                    continue
                if cstate == inspect.CORO_CLOSED:
                    continue
                with contextlib.suppress(CancelledError):
                    coro.throw(CancelledError)

            if not pending:
                break
            coros = list(pending)

        await waiter


def scope():
    return Scope()



from .._colored._ctl import (
    block_on as block_on,
    map as map,
    map_blocking as map_blocking,
    spawn as spawn,
    spawn_blocking as spawn_blocking,
    yield_now as yield_now,
)
from .._colored._events import Event as Event, Waiter as Waiter
from .._colored._scope import scope as scope
from .._deco import main as main
from .._runtime import Runtime as Runtime, new as runtime, run as run  # noqa: F401
from .time import sleep as sleep



from ._ctl import (
    block_on as block_on,
    map as map,
    map_blocking as map_blocking,
    spawn as spawn,
    spawn_blocking as spawn_blocking,
)
from ._deco import main as main
from ._events import Event as Event, Waiter as Waiter
from ._runtime import Runtime as Runtime, new as runtime, run as run  # noqa: F401
from ._scope import scope as scope
from ._tonio import __version__ as __version__
from .signals import signal_receiver as signal_receiver
from .time import sleep as sleep
